"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.TSParametricEndPoints = exports.TSStaticWebsite = void 0;
const fs_1 = require("./fs");
const commons_1 = require("./commons");
const tsrequest_1 = require("./tsrequest");
const tserrors_1 = require("./tserrors");
class TSStaticWebsite {
    constructor(uri, folder, opts) {
        if (!(0, commons_1.$length)(uri)) {
            throw `Static website url not defined for folder '${folder}'`;
        }
        if (!(0, fs_1.$isdirectory)(folder)) {
            throw `website ${uri} root folder ${folder} was not found.`;
        }
        this.uri = uri.toLowerCase();
        this.folder = folder;
        this._opts = Object.assign({}, opts);
        this._cache = new Map();
        this._cacheMemory = 0;
        this._blacklisted = new Set();
        if (!(0, commons_1.$ok)(this._opts.logger)) {
            throw 'You must specify a logger to a new TSStaticWebsite';
        }
        if (!(0, commons_1.$objectcount)(this._opts.managedTypes)) {
            this._opts.managedTypes = Object.assign({}, TSStaticWebsite.__TSServerStandardsTypes);
        }
        if (!(0, commons_1.$ok)(this._opts.maxCacheSize)) {
            this._opts.maxCacheSize = 128 * 1024 * 1024;
        }
        if (!(0, commons_1.$ok)(this._opts.maxCachedFiles)) {
            this._opts.maxCachedFiles = 10000;
        }
        if (!(0, commons_1.$ok)(this._opts.maxCachedFileSize)) {
            this._opts.maxCachedFileSize = 256 * 1024;
        }
        if (!(0, commons_1.$ok)(this._opts.maxBlacklistedFiles)) {
            this._opts.maxBlacklistedFiles = 50000;
        }
    }
    getStaticResource(uri) {
        let ret = undefined;
        let type = '';
        const lcUri = uri.toLowerCase();
        if (lcUri.startsWith(this.uri)) {
            const ext = (0, fs_1.$ext)(lcUri);
            type = (0, commons_1.$length)(ext) ? this._opts.managedTypes[ext] : '';
            if ((0, commons_1.$length)(type)) {
                const path = uri.slice(this.uri.length);
                if (!this._blacklisted.has(path)) {
                    ret = this._cache.get(path);
                    if (!(0, commons_1.$ok)(ret)) {
                        const file = (0, fs_1.$path)(this.folder, path);
                        if ((0, fs_1.$isfile)(file)) {
                            ret = (0, fs_1.$readBuffer)(file);
                            if ((0, commons_1.$ok)(ret)) {
                                this._cacheContent(path, ret);
                            }
                            else {
                                this._blacklist(path);
                            }
                        }
                        else {
                            this._blacklist(path);
                        }
                    }
                }
            }
        }
        return [ret, type];
    }
    clearCaches() {
        return __awaiter(this, void 0, void 0, function* () {
            this._cacheMemory = 0;
            this._cache.clear();
            this._blacklisted.clear();
        });
    }
    _cacheContent(path, b) {
        if (this._cache.size < this._opts.maxCachedFiles && b.length <= this._opts.maxCachedFileSize && this._cacheMemory + b.length <= this._opts.maxCacheSize) {
            this._cache.set(path, b);
            this._cacheMemory += b.length;
        }
    }
    _blacklist(path) {
        if (this._blacklisted.size < this._opts.maxBlacklistedFiles) {
            this._blacklisted.add(path);
        }
    }
}
exports.TSStaticWebsite = TSStaticWebsite;
TSStaticWebsite.__TSServerStandardsTypes = {
    'html': 'text/html',
    'htm': 'text/html',
    'css': 'text/css',
    'js': 'text/javascript',
    'svg': 'image/svg+xml',
    'svgz': 'image/svg+xml',
    'jpeg': 'image/jpeg',
    'jpg': 'image/jpg',
    'png': 'image/png',
    'gif': 'image/gif',
    'ico': 'image/x-icon',
    'pdf': 'application/pdf'
};
class TSParametricEndPoints {
    constructor(path, ep, caseInsensitive = false) {
        path = (0, commons_1.$trim)(path);
        const len = path.length;
        if (len < 2) {
            throw `End points path '${path}' is too short`;
        }
        const methods = (0, commons_1.$keys)(ep);
        if (methods.length === 0) {
            throw `End points path '${path}' has no method defined.`;
        }
        methods.forEach(m => {
            if (!(0, commons_1.$ok)(TSParametricEndPoints.validRequestMethod(m))) {
                throw `End points path '${path}' did define invalid '${m}' request method.`;
            }
        });
        this.endPoints = ep;
        this.caseInsensitive = caseInsensitive;
        // ================= automat initialization ===============
        this.uri = '/';
        this.tokens = [];
        this.depth = 1;
        let State;
        (function (State) {
            State[State["Start"] = 0] = "Start";
            State[State["Standard"] = 1] = "Standard";
            State[State["Bracket"] = 2] = "Bracket";
            State[State["Token"] = 3] = "Token";
        })(State || (State = {}));
        ;
        let state = State.Start;
        let regString = '';
        let currentToken = '';
        let constructUri = true;
        // ================= automated path analysis ===============
        for (let i = 0; i < len; i++) {
            const c = path.charAt(i);
            //console.log(`state = ${state}, char = "${c}"`) ;
            switch (state) {
                case State.Start:
                    if (c !== '/') {
                        throw `End points path '${path}' is not absolute.`;
                    }
                    state = State.Standard;
                    break;
                case State.Standard:
                    switch (c) {
                        case '{':
                            state = State.Bracket;
                            constructUri = false;
                            break;
                        case '}': throw `Misplaced '}' character in path '${path}'.`;
                        case '/':
                            if (constructUri) {
                                this.uri += c;
                            }
                            else {
                                regString += '\\/';
                            }
                            this.depth++;
                            break;
                        case '-':
                        case '$':
                        case '+':
                        case '*':
                        case '?':
                        case '.':
                        case '(':
                        case ')':
                            if (constructUri) {
                                this.uri += c;
                            }
                            else {
                                regString += '\\' + c;
                            }
                            break;
                        case '^':
                        case '\\':
                        case '[':
                        case ']':
                        case '`':
                        case '|':
                            throw `Found forbidden character '${c}' in path '${path}'.`;
                        default:
                            if (constructUri) {
                                this.uri += caseInsensitive ? c.toLowerCase() : c;
                            }
                            else {
                                regString += caseInsensitive ? c.toLowerCase() : c;
                            }
                            break;
                    }
                    break;
                case State.Bracket:
                    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')) {
                        currentToken = c;
                        state = State.Token;
                        break;
                    }
                    throw `Found formidden first character '${c}' in parametric token in path '${path}'.`;
                case State.Token:
                    if ((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9') || c === '-' || c === '_' || c === '.') {
                        currentToken += c;
                        break;
                    }
                    else if (c === '}') {
                        this.tokens.push(currentToken);
                        regString += '([a-zA-Z][.a-zA-Z0-9_\-]*)';
                        currentToken = '';
                        state = State.Standard;
                        break;
                    }
                    throw `Found formidden character '${c}' in parametric token in path '${path}'.`;
            }
        }
        if (state !== State.Standard) {
            throw `Malformed parametric path '${path}'.`;
        }
        this._regex = regString.length === 0 ? null : (caseInsensitive ? new RegExp('^' + regString + '$', 'i') : new RegExp('^' + regString + '$'));
        this._tlen = this.tokens.length;
    }
    static validRequestMethod(method) {
        if ((0, commons_1.$ok)(method)) {
            method = method.toUpperCase();
            if (TSParametricEndPoints.__verbs.includes(method)) {
                return method;
            }
        }
        return null;
    }
    get static() { return !(0, commons_1.$ok)(this._regex); }
    execute(url, method, parameters, req, res) {
        return __awaiter(this, void 0, void 0, function* () {
            const ep = this.endPoints[method];
            if (!(0, commons_1.$ok)(ep)) {
                throw new tserrors_1.TSHttpError(`Method '${method}' not implemented on url '${url.pathname}'`, tsrequest_1.Resp.NotImplemented);
            }
            yield ep(url, parameters, req, res); // this method may also throw
        });
    }
    // can return an empty string dictionary which means that our path is not parametric
    valuesFromPath(path) {
        if (this.caseInsensitive) {
            path = path.toLowerCase();
        }
        if (path.startsWith(this.uri)) {
            const rp = path.slice(this.uri.length);
            const ret = {};
            if (!rp.length && !(0, commons_1.$ok)(this._regex)) {
                return ret;
            }
            else if (rp.length && (0, commons_1.$ok)(this._regex)) {
                const m = rp.match(this._regex);
                if ((0, commons_1.$ok)(m)) {
                    if (m.length === this._tlen + 1) {
                        for (let i = 0; i < this._tlen; i++) {
                            ret[this.tokens[i]] = m[i + 1];
                        }
                    }
                    return ret;
                }
            }
        }
        return null;
    }
}
exports.TSParametricEndPoints = TSParametricEndPoints;
TSParametricEndPoints.__verbs = Object.values(tsrequest_1.Verb);
//# sourceMappingURL=tsservercomp.js.map